# Example of http headers for convenience:
# GET /some/path HTTP/1.1
# Host: localhost:8001
# Connection: keep-alive
from typing import Optional


class HttpMethod:
    OPTIONS = 'OPTIONS'
    POST = 'POST'
    GET = 'GET'


class HttpHeadersParser:
    def __init__(self, http_message: str) -> None:
        self.http_message: str = http_message
        self._detect_line_break_char()
        self.lines_of_header: list[str] = http_message.split(
            self.line_break_char,
        )

        self.method_name: str = self._get_method_name()
        self.path: str = self._get_path()
        self.query_params = self._get_query_params()
        self.body: str = self._get_body()

    def _detect_line_break_char(self) -> None:
        self.line_break_char: str = '\r'

        if '\r\n' in self.http_message:
            self.line_break_char = '\r\n'
        elif '\n' in self.http_message:
            self.line_break_char = '\n'

    def _get_method_name(self) -> str:
        """Method name is first word of first line"""
        first_line = self.lines_of_header[0]
        first_word = first_line.split(' ')[0]
        return first_word

    def _get_path_and_query_params(self) -> tuple[str, Optional[str]]:
        first_line = self.lines_of_header[0]
        second_word = first_line.split(' ')[1]
        result = second_word.split('?')
        if len(result) == 1:
            path = result[0]
            return path, None
        else:
            path, query_params = result
            return path, query_params

    def _get_path(self) -> str:
        """Path is second name in first line"""
        path, _ = self._get_path_and_query_params()
        return path

    def _get_query_params(self) -> dict:
        _, query_params_str = self._get_path_and_query_params()
        if query_params_str is None:
            return {}

        query_params = {}
        for query_param in query_params_str.split('&'):
            key, value = query_param.split('=')
            query_params[key] = value
        return query_params

    def _get_body(self) -> str:
        """Body of http message starts after two line break characters"""
        position_of_headers_end = (
            self.http_message.find(self.line_break_char * 2)
        )
        len_of_line_breaks = len(self.line_break_char * 2)
        position_of_body_starts = (
            position_of_headers_end + len_of_line_breaks
        )

        if position_of_headers_end == -1:
            return ''
        else:
            return self.http_message[position_of_body_starts:]


def create_response_headers(
    status: int, content_type: Optional[str] = None, for_options: bool = False,
) -> str:
    """Status is number, 200 or 404
    content_type examples is:
    * application/json
    * text/html.
    """
    allow_methods = 'Allow: OPTIONS, GET, POST\n'
    if content_type:
        content_type = f'Content-Type: {content_type};charset=UTF-8\n'

    return (
        f'HTTP/1.0 {status}\n'
        'Access-Control-Allow-Origin: *\n'
        'Access-Control-Allow-Methods: POST, GET, OPTIONS\n'
        'Access-Control-Allow-Headers: origin, content-type, accept\n'
        'Access-Control-Allow-Credentials: true\n'
        'Access-Control-Max-Age: 86400\n'
        f'{allow_methods if for_options else ""}'
        f'{content_type if content_type else ""}'
        f'\n'
    )



import enum
import inspect
from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar

if TYPE_CHECKING:
    from database import Base as BaseModel

T = TypeVar('T')


def get_name_of_model(model_class: 'BaseModel') -> str:
    model_name = model_class.__name__
    # remove "Orm" postfix from model name
    model_name = model_name[:-3]
    model_name = model_name.lower()
    return model_name


def get_argument_names(foo: Callable) -> tuple[str]:
    """Function return argument names of foo"""
    return tuple(dict(inspect.signature(foo).parameters).keys())


def is_special_alchemy_field(field_name: str) -> bool:
    """Function determines iF field_name is special sql alchemy field"""
    alchemy_fields = ('registry', 'metadata', 'awaitable_attrs')
    return field_name in alchemy_fields


def is_simple_alchemy_field(
    alchemy_model: 'BaseModel', field_name: str,
) -> bool:
    """Function determine if field_name is simple alchemy field.
    Simple field means str, int, etc

    If sqlalchemy if model's field has attribute "type" it means
    that field is simple"""
    return hasattr(getattr(alchemy_model, field_name), 'type')


def is_enum_alchemy_field(alchemy_model: 'BaseModel', field_name: str) -> bool:
    """Function determine if field of alchemy_model is enum"""
    field_type = getattr(alchemy_model, field_name).type.python_type
    return issubclass(field_type, enum.Enum)


def get_python_field_type_from_alchemy_field(
    alchemy_model: 'BaseModel', field_name: str,
) -> Any:
    return getattr(alchemy_model, field_name).type.python_type


def is_property_secondary_relation(
    model: 'BaseModel', attribute_name: str,
) -> bool:
    """Sqlalchemy in its models force declare one relation in two models.
    What means. For example will take model "order" and model "product".
    Every "order" may have one "product". In database, in sql when we create
    table "product", we don't write that it relates to "order", only in table
    "order" we create product_id and create foreign key, that references
    to "product" table. In this case reference from "order" to "product"
    is primary relation. Nevertheless, one "product" can reference to
    multiple orders, but it is not marked in database schema,
    therefore I say that it is secondary relation

    And this function separate primary relation in model from secondary.
    The function return True only if attribute from attribute_name
    argument is secondary relationship."""
    attribute = getattr(model, attribute_name)

    try:
        collection_class = attribute.prop.collection_class
        return bool(collection_class)
    except AttributeError:
        return False


def is_property_foreign_key(
    model_class: 'BaseModel', attribute_name: str,
) -> bool:
    attribute = getattr(model_class, attribute_name)
    try:
        foreign_keys = attribute.foreign_keys
        return bool(foreign_keys)
    except AttributeError:
        return False


def dict_set(dct: dict, path: str, value: Generic[T]) -> Generic[T]:
    keys = path.split('.')
    keys_except_last = keys[:-1]
    last_key = keys[-1]

    for key in keys_except_last:
        dct = dct.setdefault(key, {})

    dct[last_key] = value
    return dct[last_key]


def get_operation_id_for_openapi(path: str, method: str) -> str:
    return path.replace('/', '') + '_' + method.lower()



import dataclasses
from dataclasses import field, make_dataclass
from datetime import date, datetime
from typing import Any, Callable, Iterable

from marshmallow.fields import Date, DateTime, Int, Nested, Str
from marshmallow_enum import EnumField as MarshmallowEnum
from sqlalchemy.ext.asyncio import (AsyncAttrs, AsyncEngine,
                                    async_sessionmaker, create_async_engine)
from sqlalchemy.orm import DeclarativeBase

from martin_eden.base import CustomSchema
from martin_eden.utils import (get_name_of_model,
                               get_python_field_type_from_alchemy_field,
                               is_enum_alchemy_field,
                               is_property_secondary_relation,
                               is_simple_alchemy_field,
                               is_special_alchemy_field)


class Base(AsyncAttrs, DeclarativeBase):
    pass


types_map = {
    str: Str,
    int: Int,
    date: Date,
    datetime: DateTime,
}

inverse_types_map = {
    value: key for key, value in types_map.items()
}


def query_params_to_alchemy_filters(
    filters: dict, query_param: str, value: str,
) -> Any:
    """Example of query_param argument from url:
        user__first_name__like=martin"""
    model_name, field_name, method_name = query_param.split('__')

    model_class = None
    for model_class_iter in filters:
        if get_name_of_model(model_class_iter) == model_name:
            model_class = model_class_iter
    if not model_class:
        return None

    field_obj = getattr(model_class, field_name)
    method_obj = getattr(field_obj, method_name)
    if method_name == 'like':
        return method_obj(f'%{value}%')
    else:
        return method_obj(value)


class SqlAlchemyToMarshmallow(type(Base)):
    """Metaclass get sql alchemy model, creates marshmallow
    schema based on it"""

    def __new__(cls, name: str, bases: Iterable, fields: dict) -> type:
        origin_model: Base = bases[0]

        origin_model_field_names = [
            field_name for field_name in dir(origin_model)
            if not any((
                field_name.startswith('_'),
                is_special_alchemy_field(field_name),
                is_property_secondary_relation(origin_model, field_name),
            ))
        ]

        result_fields = {}
        for field_name in origin_model_field_names:
            # add simple fields: int, str, date, datetime, etc.
            if is_simple_alchemy_field(origin_model, field_name):
                python_field_type = get_python_field_type_from_alchemy_field(
                    origin_model, field_name,
                )
                if is_enum_alchemy_field(origin_model, field_name):
                    result_fields[field_name] = MarshmallowEnum(
                        python_field_type, required=False,
                    )
                else:
                    result_fields[field_name] = (
                        types_map[python_field_type](required=False)
                    )
            # add nested fields
            elif fields.get(field_name):
                result_fields[field_name] = Nested(
                    fields[field_name], required=False,
                )

        result_model = CustomSchema.from_dict(result_fields, name=name)
        return result_model


@dataclasses.dataclass
class BaseDataclass:
    pass


class MarshmallowToDataclass(type(CustomSchema)):
    def __new__(cls, name: str, bases: Iterable, fields: dict) -> type:
        origin_schema_class = bases[0]
        origin_schema = origin_schema_class()
        origin_model_fields = origin_schema.fields

        result_fields = []
        for field_name, field_type in origin_model_fields.items():
            if isinstance(field_type, Nested):
                new_type_for_dataclass = (
                    fields['__annotations__'][field_name]
                )
            elif hasattr(field_type, 'enum'):
                new_type_for_dataclass = field_type.enum
            else:
                new_type_for_dataclass = (
                    inverse_types_map[type(field_type)]
                )

            result_fields.append((
                field_name, new_type_for_dataclass, field(default=None),
            ))

        result_dataclass = make_dataclass(
            name, fields=result_fields, bases=(BaseDataclass,),
        )
        return result_dataclass


class DataBase:
    def __init__(self) -> None:
        self.engine: AsyncEngine = create_async_engine(
            'postgresql+asyncpg://alexander.bezgin:123@localhost/framework',
            echo=True,
        )
        self.create_session: Callable = async_sessionmaker(self.engine)



from typing import Any, ParamSpecArgs, ParamSpecKwargs

from marshmallow import Schema
from marshmallow.decorators import post_dump
from marshmallow_jsonschema import JSONSchema


class Controller:
    """The class needs only as type hint"""
    request_schema: Schema
    response_schema: Schema
    query_params: dict

    def __call__(
        self, *args: ParamSpecArgs, **kwargs: ParamSpecKwargs,
    ) -> None:
        pass

    def __await__(self):
        pass


class CustomSchema(Schema):
    def __init__(
        self,
        *args: ParamSpecArgs,
        json_schema_name: str = None,
        **kwargs: ParamSpecKwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.json_schema_name = json_schema_name
        self.__name__ = json_schema_name


class CustomJsonSchema(JSONSchema):
    @post_dump
    def wrap(self, data: dict, **_) -> dict[str, Any]:
        """Wrap this with the root schema definitions."""
        if self.nested:  # no need to wrap, will be in outer defs
            return data

        schema_name = self.obj.json_schema_name

        data["additionalProperties"] = False

        self._nested_schema_classes[schema_name] = data
        root = {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "definitions": self._nested_schema_classes,
            "$ref": f"#/definitions/{schema_name}",
        }
        return root